#!/usr/bin/python

"""
Here are some relevant C lines from my Linux 2.4.19 include files.


asm/posix_types.h:
    typedef long            __kernel_time_t;
    typedef long          __kernel_suseconds_t;

linux/types.h:
    typedef __kernel_time_t         time_t;
    typedef __kernel_suseconds_t      suseconds_t;

linux/time.h:
    struct timeval {
            time_t          tv_sec;         /* seconds */
            suseconds_t     tv_usec;        /* microseconds */
    };

linux/input.h:
    #define EVIOCGNAME(len)           _IOC(_IOC_READ, 'E', 0x06, len)         /* get device name */

    struct input_event {
            struct timeval time;
            unsigned short type;
            unsigned short code;
            unsigned int value;
    };

asm/ioctl.h:
    #define _IOC_NRBITS     8
    #define _IOC_TYPEBITS   8
    #define _IOC_SIZEBITS   14
    #define _IOC_DIRBITS    2

    #define _IOC_NRMASK     ((1 << _IOC_NRBITS)-1)
    #define _IOC_TYPEMASK   ((1 << _IOC_TYPEBITS)-1)
    #define _IOC_SIZEMASK   ((1 << _IOC_SIZEBITS)-1)
    #define _IOC_DIRMASK    ((1 << _IOC_DIRBITS)-1)

    #define _IOC_NRSHIFT    0
    #define _IOC_TYPESHIFT  (_IOC_NRSHIFT+_IOC_NRBITS)
    #define _IOC_SIZESHIFT  (_IOC_TYPESHIFT+_IOC_TYPEBITS)
    #define _IOC_DIRSHIFT   (_IOC_SIZESHIFT+_IOC_SIZEBITS)

    #define _IOC(dir,type,nr,size) \
            (((dir)  << _IOC_DIRSHIFT) | \
             ((type) << _IOC_TYPESHIFT) | \
             ((nr)   << _IOC_NRSHIFT) | \
             ((size) << _IOC_SIZESHIFT))

    #define _IOC_READ       2U

    #define _IOR(type,nr,size)      _IOC(_IOC_READ,(type),(nr),sizeof(size))
"""

import struct,fcntl,os,string,select, sys

class IOR:
    """a callable IOCTL instance which you can call with device files."""
    def __init__(self,type_,nr,size):
        """set up an ioctl for the given ioctl type (the second arg to
        _IOR macro)"""
        NRSHIFT    =0
        TYPESHIFT  =8
        SIZESHIFT  =16
        DIRSHIFT   =30
        READ=2L
        self.op = (READ<<DIRSHIFT) | (ord(type_)<<TYPESHIFT) | (nr<<NRSHIFT) | (size<<SIZESHIFT)
        self.op = self.op - ((sys.maxint + 1) * 2) # bogus (but WFM)

        self.size=size
        
    def __call__(self,fileno,arg=None):
        """Run the ioctl on the file and return the result string.
        The return will always be self.size chars, I think.  If you
        were expecting a string, you'd probably want to rstrip()
        it. There's also probably going to be an extra \0, so include
        that in your rstrip() list.
        """
        if arg is None:
            arg=" "*self.size
        return fcntl.ioctl(fileno, self.op, arg)

def finddevice():
    """walk through the /dev/input/* devices, returning the first one
    (as an opened file) that has the correct device name
    string. Raises IOError if no good device is found"""

    EVIOCGNAME = IOR('E', 0x06, 255)

    shuttleproname='Contour Design ShuttlePRO'
    dev=None
    
    for devname in [os.path.join("/dev/input",x) for x in os.listdir("/dev/input")]:
        try:
            dev=open(devname,"rb")

            devicename=EVIOCGNAME(dev.fileno()).rstrip(string.whitespace+"\0")
            if devicename==shuttleproname:
                print "found %s at %s" % (shuttleproname,devname)
                break
        except IOError:
            # probably 'Device not found' or some other reason to skip this device
            dev=None

    if dev is None:
        raise IOError("can't find a ShuttlePRO")
    
    return dev



# changed final int to signed, since the jog value is signed (and the
# rest don't get big enough to matter)
input_event_struct="llHHi" 
input_event_struct_len=struct.calcsize(input_event_struct)


class Event:
    """shuttlepro event, returned by Shuttlepro.readevent.

    event will have a 'type' attribute, which is the string 'button',
    'jog', or 'shuttle'. see the __str__ for the other fields for each
    type.

    """
    def __init__(self):
        self.type = "undefined"
        self.position = self.dir = None
    def knownevent(self):
        return self.type not in ["undefined","unknown"]
    def __str__(self):
        if self.type=='button':
            return "button number %s %s" % (self.number,self.state)
        elif self.type=='jog':
            return "jog %d" % (self.dir)
        elif self.type=='shuttle':
            return "shuttle %d" % self.position
        elif self.type=='unknown':
            return "unknown: type=%s, code=%s, value=%s" % (self.evtype,self.code,self.value)
        else:
            return "no event"


class Shuttlepro:
    def __init__(self):
        """if no shuttle is found, this object stays 'dead'. you can
        still call readevent on it, and it'll always return a blank
        Event """

        try:
            self.dev=finddevice()
        except IOError:
            # make dead object
            self.readevent=lambda *args,**kw: Event()
            
        self.prevjog=None

    def readevent(self,block=1):
        """this might return a null event on the first jog event, since that one has no data"""
        if not block:
            iready,oready,eready=select.select([self.dev],[],[],0)
            if self.dev not in iready:
                return Event()
        input_event=self.dev.read(input_event_struct_len)
        sec,usec,evtype,code,value=struct.unpack(input_event_struct,input_event)

        ev=Event()

        if evtype==1:
            # button event
            ev.type="button"
            ev.number=code-256
            ev.state=['up','down'][value]

        elif (evtype,code)==(2,7):
            ev.type="jog"

            value=value-1 # values go 1..255. now they're 0..254
            
            if self.prevjog==None:
                # There's nothing interesting about the first seen value
                # from jog. 
                self.prevjog=value

            # If we know the previous jog value, we can calculate
            # the move, which is -1, 0, or 1. There are zeros since
            # the device reports its jog setting about every 3
            # seconds. A dir beyond 1 is possible, although I haven't
            # seen it.

            ev.dir=(value-self.prevjog)
            if ev.dir>100:
                ev.dir=ev.dir-255 # handle the wraparound 
            if ev.dir<-100:
                ev.dir=ev.dir+255

            self.prevjog=value

        elif (evtype,code)==(2,8):
            ev.type="shuttle"
            ev.position=value

        else:
            ev.type="unknown"
            ev.evtype=evtype
            ev.code=code
            ev.value=value

        return ev
    

if __name__=='__main__':
    sp=Shuttlepro()
    while 1:
        ev=sp.readevent(block=1)
        if ev.knownevent() and (ev.type,ev.dir) != ('jog',0):
            print str(ev)
